// -*- mode: cpp; mode: fold -*-
// Description								/*{{{*/
// $Id: tar.cc 96 2006-11-27 18:56:22Z jpra $
/* ######################################################################

   Tar - A wrapper class for the GNU tar command.

   AUTHOR: David Salvador Pinheiro <david.pinheiro@caixamagica.pt>
   
   ##################################################################### */
									/*}}}*/
// Include Files							/*{{{*/
#ifdef __GNUG__
#pragma implementation "apt-pkg/tar.h"
#endif

#include <apt-pkg/tar.h>

#include <stdio.h>
#include <errno.h>
#include <string.h>
									/*}}}*/

#define RD_BUFSZ 512

const char *Tar::_ClassId = "$Id: tar.cc 96 2006-11-27 18:56:22Z jpra $";

Tar::Tar()
{
}

Tar::~Tar()
{
}

bool Tar::Create(const char *FileName,vector<string> &FileList, CompressionType Type, ostream *out ,ostream * error)
{   
   FILE *PD;
   string cmd("tar cf");
   char S[RD_BUFSZ];
   
   switch(Type)
   {
      case NONE:
      break;

      case BZ2:
      cmd.append("j");
      break;

      case GZIP:
      cmd.append("z");
      break;
   }

   cmd.append(" ").append(FileName).append(" ");
   
   for (vector<string>::iterator I = FileList.begin(); I != FileList.end(); I++)
      cmd.append(*I).append(" ");
   
   PD = popen(cmd.c_str(),"r");
   
   if(PD == NULL)
   {
      if(error != NULL)
         *error << strerror(errno);

      return false;
   }
   
   while(fgets(S,sizeof S, PD)!= NULL)
   {
      if(out != NULL)
	 *out << S;
   }
   
   pclose(PD);
   return true;
}

bool Tar::Extract(const char * FileName, CompressionType Type, ostream *out, ostream *error)
{
   FILE *PD;
   string cmd("tar xf");
   char S[RD_BUFSZ];
   
   switch(Type)
   {
      case NONE:
      break;

      case BZ2:
      cmd.append("j");
      break;

      case GZIP:
      cmd.append("z");
      break;
   }

   cmd.append(" ").append(FileName).append(" ");
   
   PD = popen(cmd.c_str(),"r");
   
   if(PD == NULL)
   {
      if(error != NULL)
         *error << strerror(errno);

      return false;
   }
   
   while(fgets(S,sizeof S, PD)!= NULL)
   {
      if(out != NULL)
	 *out << S;
   }
   
   pclose(PD);
   return true;
}


// vim:sts=3:sw=3
