// -*- mode: c++; mode: fold -*-
// Description								/*{{{*/
// $Id: tagfile.h,v 1.20 2003/05/19 17:13:57 doogie Exp $
/* ######################################################################

   Fast scanner for Urpm Synthesis lists: 
   based on Debian apt pkgTag* classes.
   
   The parser reads the file and provides methods to step linearly
   over it or to jump to a pre-recorded start point and read that record.
   
   A second class is used to perform pre-parsing of the record. It works
   by indexing the start of each header field and providing lookup 
   functions for header fields.
   
   ##################################################################### */
									/*}}}*/
#ifndef PKGLIB_URPMLISTFILE_H
#define PKGLIB_URPMLISTFILE_H

#include <apt-pkg/fileutl.h>
#include <stdio.h>
    
class UrpmListSection
{
   const char *Section;
   const char *Stop;
   
   // We have a limit of 256 tags per section.
   unsigned int Indexes[256];
   unsigned int AlphaIndexes[0x100];
   
   unsigned int TagCount;
     
   public:
   
   inline bool operator ==(const UrpmListSection &rhs) {return Section == rhs.Section;}
   inline bool operator !=(const UrpmListSection &rhs) {return Section != rhs.Section;}
   
   bool Find(const char *Tag,const char *&Start, const char *&End) const;
   bool Find(const char *Tag,unsigned &Pos) const;
   string FindS(const char *Tag) const;
   string FindSubS(const char *Tag, const char *SubTag) const;
   signed int FindI(const char *Tag,signed long Default = 0) const ;
   bool Scan(const char *Start,unsigned long MaxLength);
   inline unsigned long size() const {return Stop - Section;};
   void Trim();
   
   inline unsigned int Count() const {return TagCount;};
   inline void Get(const char *&Start,const char *&Stop,unsigned int I) const
                   {Start = Section + Indexes[I]; Stop = Section + Indexes[I+1];}
	    
   inline void GetSection(const char *&Start,const char *&Stop) const
   {
      Start = Section;
      Stop = this->Stop;
   }
   
   UrpmListSection() : Section(0), Stop(0) {}
};

class UrpmListFile
{
   FileFd &Fd;
   char *Buffer;
   char *Start;
   char *End;
   bool Done;
   off_t iOffset;
   unsigned long Size;
   
   bool Fill();
   
   public:

   bool Step(UrpmListSection &Section);
   inline off_t Offset() {return iOffset;}
   bool Jump(UrpmListSection &Tag,off_t Offset);

   UrpmListFile(FileFd *F,size_t Size = 32*1024);
   ~UrpmListFile();
};

#endif
